#!/usr/bin/env python3

import torch
import torch.nn as nn
import torchvision
import torchvision.transforms as transforms
import numpy as np
import shutil
import os

# hyper-parameters
num_classes = 10
batch_size = 100
learning_rate = 5e-4

# channel sizes
first_channel = 1
second_channel = 16
third_channel = 32

class ConvolutionalNetwork(nn.Module):
    def __init__(self, num_of_classes = num_classes):
        super(ConvolutionalNetwork, self).__init__()
        self.conv1 = nn.Conv2d(in_channels=first_channel, out_channels=second_channel,
                               kernel_size=5, padding=2)
        self.batchnorm1 = nn.BatchNorm2d(num_features=second_channel)
        self.relu1 = nn.ReLU()
        self.maxpool1 = nn.MaxPool2d(kernel_size=2, stride=2)
        self.conv2 = nn.Conv2d(in_channels=second_channel, out_channels=third_channel,
                               kernel_size=5, padding=2)
        self.batchnorm2 = nn.BatchNorm2d(num_features=third_channel)
        self.relu2 = nn.ReLU()
        self.maxpool2 = nn.MaxPool2d(kernel_size=2, stride=2)
        self.fc = nn.Linear(in_features=7*7*32, out_features=num_classes)
    def forward(self, x):
        x = self.conv1(x)
        x = self.batchnorm1(x)
        x = self.relu1(x)
        x = self.maxpool1(x)
        x = self.conv2(x)
        x = self.batchnorm2(x)
        x = self.relu2(x)
        x = self.maxpool2(x)
        x = x.reshape(-1, 7*7*32)
        x = self.fc(x)
        return x

    
epoch_to_save = [0,1,3,7,15,31,63]

if __name__ == "__main__":
    # identify the device
    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

    # load the MNIST dataset
    train_dataset = torchvision.datasets.MNIST(root='data/',
                                               train=True,
                                               transform=transforms.ToTensor(),
                                               download=True)

    # define the data loader
    train_loader = torch.utils.data.DataLoader(dataset=train_dataset,
                                               batch_size=batch_size,
                                               shuffle=True)

    # define model, criterion and optimizer
    model = ConvolutionalNetwork().to(device)
    criterion = nn.CrossEntropyLoss()
    optimizer = torch.optim.Adam(model.parameters(), lr=learning_rate)

    # clean up the environment
    if os.path.isdir('model'):
        shutil.rmtree('model');
    os.mkdir('model');

    # train the model
    for epoch in range(max(epoch_to_save)+1):
        for index, (images, labels) in enumerate(train_loader):
            images = images.to(device)

            labels = labels.to(device)

            # forward pass
            outputs = model(images)
            loss = criterion(outputs, labels)

            # backward pass and optimize
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()

            if (index+1) % 100 == 0:
                print('Epoch [{}/{}], Step [{}/{}], Loss: {:.4f}'
                      .format(epoch + 1, max(epoch_to_save)+1, index+1, len(train_loader), loss.item()))

        # save the model
        if epoch in epoch_to_save:
            torch.save(model, "model/e{}.pt".format(epoch))